#!/usr/bin/env python

import sys, os, select, socket, time, traceback
import string, re

import IRCEvents


class IRCError (Exception):
   def __init__ (self, value=None):
      self.value = value

   def __str__ (self):
      return str (self.value)



class IRCClient:
   def __init__ (self, host, port):
      self.server = host
      self.serverport = port
      self.socket = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
      self.socket.connect ((self.server, self.serverport))

      self.loggedin = 0
      self.nick = None
      self.channels = []
      
      self.debuglevel = 0


   def debug (self, message, debuglevel):
      if debuglevel <= self.debuglevel:
         sys.stderr.write ("DEBUG" + " " * debuglevel + message + "\n")

      
   def send_raw (self, command):
      data = command + "\r\n"
      self.debug ("Attempting: " + repr (command), 11)
      if self.loggedin:
         while data:
            bytes_sent = self.socket.send (data)
            data = data[bytes_sent:]
      else:
         raise IRCError, "Not yet logged in!"
      self.debug ("sent: " + repr (command), 10)


   def get_raw (self):
      buffer = ''
      while select.select ([self.socket], [], [], 20)[0]:
         x = self.socket.recv(1)
         buffer = buffer + x
         if buffer and buffer[-1] == '\n':
            buffer = buffer[:-1]
            if buffer[-1] == '\r':
               buffer = buffer[:-1]
            break
      self.debug ("got: " + repr (buffer), 10)
      return buffer


   def testmessage (self, timeout=0):
      if select.select ([self.socket], [], [], timeout)[0]:
         return 1
      else:
         return None
   
      
   def getmessage (self):
      retval = IRCEvents.IRCEvent (self.get_raw (), self)
      self.debug (str (retval), 11)
      # answer PING events.
      if retval.type == "PING":
         self.send_raw ("PONG :%s" % retval.optional)
      return retval

      
   def login (self, nick, password = None,
                          username = "anon", 
                          realname = "Anonymous", 
                          hostname = "anonhost.nowhere.foo",
                          servername = None):
      if not username:
         username = nick
      if not servername:
         servername = self.server

      self.nick = nick
      self.loggedin = 1

      if password:
         self.send_raw ("PASS " + password)
      self.send_raw ("USER %s %s %s :%s" % (username, hostname,
                                            servername, realname))
      self.send_raw ("NICK " + nick)

      
   def join (self, channel, password = None):
      self.channels.append (channel)
      if password:
         self.send_raw ("JOIN %s %s" % (channel, password))
      else:
         self.send_raw ("JOIN %s" % channel)
         

   def privmsg (self, target, message):
      self.send_raw ("PRIVMSG %s :%s" % (target, message))


   def logout (self, reason):
      self.send_raw ("QUIT :%s" % (reason))



   # Higher level functions

   def react (self, msg, text, action=0):
      if action:
         self.privmsg (msg.replyto, "\x01ACTION %s\x01" % (text % msg))
      else:
         self.privmsg (msg.replyto, text % msg)


   def reply (self, msg, text):
      self.privmsg (msg.replyto, msg.nick + ": " + (text % msg))




class IRCBot (IRCClient):
   def __init__ (self, server, port=6667):
      IRCClient.__init__ (self, server, port)
      self.commands = {}
      self.actions  = []
      self.schedule = []
      self.idlelist = []
      self.modules = {}
      self.modulepath = [ os.getcwd() + "/modules" ]


   def mainloop (self):
      self.quitmessage = ''

      try:
         while not self.quitmessage:
            if self.testmessage ():
               message = self.getmessage ()
               self.dispatch_message (message)
     
            self.do_timeout ()
            if not self.do_idle ():
               time.sleep (0.1)
         self.logout (self.quitmessage)

      except KeyboardInterrupt:
         self.logout ("Eeeeek! CTRL-C!")


   def stop (self, quitmessage=""):
      if not quitmessage:
         quitmessage = "bye!"
      self.quitmessage = quitmessage


   def load_module (self, name, refresh=0):
      bak_syspath = sys.path[:]
      sys.path += self.modulepath

      try:
         if refresh and self.modules.has_key (name):
            module = reload (self.modules[name])
         else:
            module = __import__ (name)

         self.modules [name] = module
      except:
         print "------------------------------------"
         print "%sLoading of %s failed!" % ("Re" * refresh, name)
         traceback.print_exc ()
         print "------------------------------------"
         module = None

      sys.path = bak_syspath
      self.rebuild_action_list ()

      return module


   def unload_module (self, name):
      if self.modules.has_key (name):
         del self.modules[name]
         self.rebuild_action_list ()
      

   def reload_all (self):
      ret = 1

      for name in self.modules.keys ():
         if not self.load_module (name, 1):
            ret = 0

      return ret


   def rebuild_action_list (self):
      self.commands = {}
      self.actions = []
      self.idlelist = []
    
      for name in self.modules.keys ():
         try:
            bak_actions = self.actions[:]
            bak_idlelist = self.idlelist[:]
            bak_commands = self.commands.copy ()
            module = self.modules[name]
            def dummy (x): return []
            self.actions += getattr (module, "get_handlers", dummy) (self)
            self.idlelist += getattr (module, "get_idlefuncs", dummy) (self)
            def dummy2 (x): return {}
            self.commands.update (getattr (module, "get_commands", dummy2) (self))
         except:
            print "------------------------------------"
            print "Getting the handlers of %s failed, activating backup!" % name
            traceback.print_exc ()
            print "------------------------------------"
            del self.modules[name]
            self.actions  = bak_actions 
            self.idlelist = bak_idlelist
            self.commands = bak_commands

      self.actions.sort ()
      self.actions.reverse ()


   def invoke_command (self, message, commandstring):
      todo = re.search ("^\s*(\w+)\s*(.*)", commandstring)

      if todo:
         command = todo.group (1).lower ()
         arguments = todo.group (2)
         if not self.commands.has_key (command):
            return None

         try:
            return self.commands[command] (message, arguments)
         except:
            print "--------------------------"
            print command, "failed!"
            print commandstring
            traceback.print_exc()
            print "--------------------------"

      return None


   def dispatch_message (self, message):
      # first try to invoke a simple command.
      if (message.type == 'PRIVMSG' and message.command):
         if self.invoke_command (message, message.command):
            return

      # action:  (priority, eventmask, function, args, kwargs)
      for command in self.actions:
         try:
            if command [1].match (message):
               if apply (command[2], (message, ) + command[3], command[4]):
                  break
         except:
            print "--------------------------"
            print command, "failed!"
            print message
            traceback.print_exc()
            print "--------------------------"


   def do_timeout (self):
      if len (self.schedule) >= 1:
         t = time.time ()
         if self.schedule[0][0] <= t:
            command = self.schedule[0]
            try: 
               apply (command[1], command[2], command[3])
            except:
               print "--------------------------"
               print command, "failed!"
               traceback.print_exc()
               print "--------------------------"

            self.schedule = self.schedule[1:]


   def do_idle (self):
      ret = 0
      if len (self.idlelist) >= 1:
         command = self.idlelist[0]
         try: 
            ret = apply (command[0], (self, ) + command[1], command[2])
         except Exception, v:
            print "--------------------------"
            print command, "failed!"
            traceback.print_exc()
            print "--------------------------"

         self.idlelist = self.idlelist [1:]
         self.idlelist.append (command)
      return ret


   def add_timeout (self, time, function, *args, **kwargs):
      self.schedule.append ( (time, function, args, kwargs) )
      self.schedule.sort ()


