#!/usr/bin/env python

import re, string, pprint


def create_ctcp (regex, function, *args, **kwargs):
   eventmask = IRCEventMask (type = "PRIVMSG", ctcp = regex)
   return apply (create_handler, (0, eventmask, function) + args, kwargs)


def create_command (regex, function, *args, **kwargs):
   eventmask = IRCEventMask (type = "PRIVMSG", command = regex)
   return apply (create_handler, (0, eventmask, function) + args, kwargs)


def create_reaction (regex, function, *args, **kwargs):
   eventmask = IRCEventMask (type = "PRIVMSG", text = regex)
   return apply (create_handler, (0, eventmask, function) + args, kwargs)


def create_handler (priority, eventmask, function, *args, **kwargs):
   return ((priority, eventmask, function, args, kwargs))

   


class IRCEventMask:
   def __init__ (self, **filters):
      self.filters = {}
      for i in filters.keys():
         self.filters[i] = re.compile (str (filters[i]))


   def match (self, event):
      for i in self.filters.keys():
         if ( not hasattr (event, i) or getattr (event, i) == None or
              not self.filters[i].search (str (getattr (event, i)))):
            return None

      return 1




class IRCEvent:
   adressee_re = re.compile ("^ *(\S+) *[:,] *(.*)")

   def __init__ (self, messagestring, client):
      self.msgstring = messagestring
      self.client = client

      self.type = None
      self.args = []
      self.optional = None
      self.error = None

      self.sender = None
      self.nick = None
      self.username = None
      self.hostname = None
      self.server = None

      # split the message in three parts: (optional) origin, command,
      # (optional) argument with spaces.
      origin = None
      if messagestring[0] == ":":
         origin, messagestring = string.split (messagestring[1:], " ", 1)

      if ":" in messagestring:
         messagestring, self.optional = string.split (messagestring, ":", 1)

      message = string.split (messagestring)
      self.args = message [1:]

      # parsing the (optional) origin of a message

      self.sender = origin
      if origin:
         if "@" in origin:
            origin, self.hostname = string.split (origin, "@", 1)
            if "!" in origin:
               self.nick, self.username = string.split (origin, "!", 1)
            else:
               self.nick = origin
         else:
            if "." in origin:
               self.server = origin
            else:
               self.nick = origin
         
      # parsing the type of the message

      type = replydict.get (message[0])
      number = None
      try:
         number = int (message[0]) 
         if number >= 400 and number < 600:
            self.error = 1
      except ValueError:
         self.error = None

      if type:
         self.type = type
      else:
         if number:
            self.client.debug ("UNKNOWN PROTOCOL: %s %s" % (message[0], repr (self.msgstring)), 7)
         self.type = message[0]

      self.client.debug ("-------------------------------", 7)
      self.client.debug ("FROM: %s\n                  (%s ! %s @ %s  //  %s)"
                            % (self.sender,
                               self.nick,
                               self.username,
                               self.hostname,
                               self.server),
                         7)
      self.client.debug ("TYPE:     %s" % self.type, 7)
      self.client.debug ("ARGS:     %s" % repr (message[1:]), 7)
      self.client.debug ("OPTIONAL: %s" % repr (self.optional), 7)

      parser = getattr (self, "parse_" + string.lower (self.type), None)
      if parser:
         parser (message [1:], self.optional)
      else:
         self.client.debug ("no special case handling.", 8)


   def __str__ (self):
      return ("\n" + pprint.pformat (self.__dict__))


   def __getitem__ (self, key):
      return getattr (self, key)


   def get (self, key, default=None):
      return getattr (self, key, default)


   def parse_privmsg (self, args, optional):
      self.adressee = None
      self.command  = None
      self.replyto  = None
      self.ctcp     = None

      self.rcpt = args[0]
      self.text = optional

      if self.text and self.text[0] == self.text[-1] and self.text[0] == '\x01':
         self.ctcp = string.split (self.text[1:-1], " ", 1)[0]

      candidate = IRCEvent.adressee_re.match (self.text)

      if self.rcpt[0] == "#":
         self.replyto = self.rcpt
         if candidate:
            self.adressee = candidate.group (1)
            if string.lower (self.adressee) == string.lower (self.client.nick):
               self.command  = candidate.group (2)
      else:
         self.replyto = self.nick
         self.adressee = self.rcpt
         if string.lower (self.adressee) == string.lower (self.client.nick):
            self.command  = string.lstrip (self.text)
        



# boring stuff below...

replydict = {

  # Command responses

  '001' : 'RPL_WELCOME',
  '002' : 'RPL_YOURHOST',
  '003' : 'RPL_CREATED',
  '004' : 'RPL_MYINFO',
  '005' : 'RPL_BOUNCE',

  '302' : 'RPL_USERHOST',
  '303' : 'RPL_ISON',
  '301' : 'RPL_AWAY',
  '305' : 'RPL_UNAWAY',
  '306' : 'RPL_NOWAWAY',
  '311' : 'RPL_WHOISUSER',
  '312' : 'RPL_WHOISSERVER',
  '313' : 'RPL_WHOISOPERATOR',
  '317' : 'RPL_WHOISIDLE',
  '318' : 'RPL_ENDOFWHOIS',
  '319' : 'RPL_WHOISCHANNELS',
  '314' : 'RPL_WHOWASUSER',
  '369' : 'RPL_ENDOFWHOWAS',
  '321' : 'RPL_LISTSTART',
  '322' : 'RPL_LIST',
  '323' : 'RPL_LISTEND',
  '325' : 'RPL_UNIQOPIS',
  '324' : 'RPL_CHANNELMODEIS',
  '331' : 'RPL_NOTOPIC',
  '332' : 'RPL_TOPIC',
  '341' : 'RPL_INVITING',
  '342' : 'RPL_SUMMONING',
  '346' : 'RPL_INVITELIST',
  '347' : 'RPL_ENDOFINVITELIST',
  '348' : 'RPL_EXCEPTLIST',
  '349' : 'RPL_ENDOFEXCEPTLIST',
  '351' : 'RPL_VERSION',
  '352' : 'RPL_WHOREPLY',
  '315' : 'RPL_ENDOFWHO',
  '353' : 'RPL_NAMREPLY',
  '366' : 'RPL_ENDOFNAMES',
  '364' : 'RPL_LINKS',
  '365' : 'RPL_ENDOFLINKS',
  '367' : 'RPL_BANLIST',
  '368' : 'RPL_ENDOFBANLIST',
  '371' : 'RPL_INFO',
  '374' : 'RPL_ENDOFINFO',
  '375' : 'RPL_MOTDSTART',
  '372' : 'RPL_MOTD',
  '376' : 'RPL_ENDOFMOTD',
  '381' : 'RPL_YOUREOPER',
  '382' : 'RPL_REHASHING',
  '383' : 'RPL_YOURESERVICE',
  '391' : 'RPL_TIME',
  '392' : 'RPL_USERSSTART',
  '393' : 'RPL_USERS',
  '394' : 'RPL_ENDOFUSERS',
  '395' : 'RPL_NOUSERS',
  '200' : 'RPL_TRACELINK',
  '201' : 'RPL_TRACECONNECTING',
  '202' : 'RPL_TRACEHANDSHAKE',
  '203' : 'RPL_TRACEUNKNOWN',
  '204' : 'RPL_TRACEOPERATOR',
  '205' : 'RPL_TRACEUSER',
  '206' : 'RPL_TRACESERVER',
  '207' : 'RPL_TRACESERVICE',
  '208' : 'RPL_TRACENEWTYPE',
  '209' : 'RPL_TRACECLASS',
  '210' : 'RPL_TRACERECONNECT',
  '261' : 'RPL_TRACELOG',
  '262' : 'RPL_TRACEEND',
  '211' : 'RPL_STATSLINKINFO',
  '212' : 'RPL_STATSCOMMANDS',
  '219' : 'RPL_ENDOFSTATS',
  '242' : 'RPL_STATSUPTIME',
  '243' : 'RPL_STATSOLINE',
  '221' : 'RPL_UMODEIS',
  '234' : 'RPL_SERVLIST',
  '235' : 'RPL_SERVLISTEND',
  '251' : 'RPL_LUSERCLIENT',
  '252' : 'RPL_LUSEROP',
  '253' : 'RPL_LUSERUNKNOWN',
  '254' : 'RPL_LUSERCHANNELS',
  '255' : 'RPL_LUSERME',
  '256' : 'RPL_ADMINME',
  '257' : 'RPL_ADMINLOC1',
  '258' : 'RPL_ADMINLOC2',
  '259' : 'RPL_ADMINEMAIL',
  '263' : 'RPL_TRYAGAIN',


  # Errors

  '401' : 'ERR_NOSUCHNICK',
  '402' : 'ERR_NOSUCHSERVER',
  '403' : 'ERR_NOSUCHCHANNEL',
  '404' : 'ERR_CANNOTSENDTOCHAN',
  '405' : 'ERR_TOOMANYCHANNELS',
  '406' : 'ERR_WASNOSUCHNICK',
  '407' : 'ERR_TOOMANYTARGETS',
  '408' : 'ERR_NOSUCHSERVICE',
  '409' : 'ERR_NOORIGIN',
  '411' : 'ERR_NORECIPIENT',
  '412' : 'ERR_NOTEXTTOSEND',
  '413' : 'ERR_NOTOPLEVEL',
  '414' : 'ERR_WILDTOPLEVEL',
  '415' : 'ERR_BADMASK',
  '421' : 'ERR_UNKNOWNCOMMAND',
  '422' : 'ERR_NOMOTD',
  '423' : 'ERR_NOADMININFO',
  '424' : 'ERR_FILEERROR',
  '431' : 'ERR_NONICKNAMEGIVEN',
  '432' : 'ERR_ERRONEUSNICKNAME',
  '433' : 'ERR_NICKNAMEINUSE',
  '436' : 'ERR_NICKCOLLISION',
  '437' : 'ERR_UNAVAILRESOURCE',
  '441' : 'ERR_USERNOTINCHANNEL',
  '442' : 'ERR_NOTONCHANNEL',
  '443' : 'ERR_USERONCHANNEL',
  '444' : 'ERR_NOLOGIN',
  '445' : 'ERR_SUMMONDISABLED',
  '446' : 'ERR_USERSDISABLED',
  '451' : 'ERR_NOTREGISTERED',
  '461' : 'ERR_NEEDMOREPARAMS',
  '462' : 'ERR_ALREADYREGISTRED',
  '463' : 'ERR_NOPERMFORHOST',
  '464' : 'ERR_PASSWDMISMATCH',
  '465' : 'ERR_YOUREBANNEDCREEP',
  '466' : 'ERR_YOUWILLBEBANNED',
  '467' : 'ERR_KEYSET',
  '471' : 'ERR_CHANNELISFULL',
  '472' : 'ERR_UNKNOWNMODE',
  '473' : 'ERR_INVITEONLYCHAN',
  '474' : 'ERR_BANNEDFROMCHAN',
  '475' : 'ERR_BADCHANNELKEY',
  '476' : 'ERR_BADCHANMASK',
  '477' : 'ERR_NOCHANMODES',
  '478' : 'ERR_BANLISTFULL',
  '481' : 'ERR_NOPRIVILEGES',
  '482' : 'ERR_CHANOPRIVSNEEDED',
  '483' : 'ERR_CANTKILLSERVER',
  '484' : 'ERR_RESTRICTED',
  '485' : 'ERR_UNIQOPPRIVSNEEDED',
  '491' : 'ERR_NOOPERHOST',
  '501' : 'ERR_UMODEUNKNOWNFLAG',
  '502' : 'ERR_USERSDONTMATCH',


  # reserved, partially in use.

  '231' : 'RPL_SERVICEINFO',
  '232' : 'RPL_ENDOFSERVICES',
  '233' : 'RPL_SERVICE',

  '300' : 'RPL_NONE',
  '316' : 'RPL_WHOISCHANOP',
  '361' : 'RPL_KILLDONE',
  '362' : 'RPL_CLOSING',
  '363' : 'RPL_CLOSEEND',
  '373' : 'RPL_INFOSTART',
  '384' : 'RPL_MYPORTIS',

  '213' : 'RPL_STATSCLINE',
  '214' : 'RPL_STATSNLINE',
  '215' : 'RPL_STATSILINE',
  '216' : 'RPL_STATSKLINE',
  '217' : 'RPL_STATSQLINE',
  '218' : 'RPL_STATSYLINE',
  '240' : 'RPL_STATSVLINE',
  '241' : 'RPL_STATSLLINE',
  '244' : 'RPL_STATSHLINE',
  '244' : 'RPL_STATSSLINE',
  '246' : 'RPL_STATSPING',
  '247' : 'RPL_STATSBLINE',
  '250' : 'RPL_STATSDLINE',

  '492' : 'ERR_NOSERVICEHOST',


  # not defined in rfc 2812

  '265' : 'RPL_LOCALUSERS',
  '266' : 'RPL_GLOBALUSERS',
  '333' : 'RPL_TOPICWHOTIME',

}
